import requests
from .cache import fetch_result_dbpedia, save_query_results_dbpedia
wikidata_endpoint = 'https://query.wikidata.org/bigdata/namespace/wdq/sparql'
dbpedia_endpoint = 'https://dbpedia.org/sparql'


def do_query_helper(endpoint, query):
    cached_results = fetch_result_dbpedia(query)
    if len(cached_results) > 0:
        print('Using cached query results')
        return cached_results # eval(cached_results)

    params = {'query': query, 'format': 'json'}
    response = requests.get(endpoint, params=params)
    json_data = response.json()
    vars = json_data['head']['vars']
    results = json_data['results']
    if 'bindings' in results:
        bindings = results['bindings']
        qr = [[[var, binding[var]['value']] for var in vars] for binding in bindings]
        save_query_results_dbpedia(query, qr)
        return qr
    return []


def wikidata_sparql(query):
    return do_query_helper(wikidata_endpoint, query)


def dbpedia_sparql(query):
    return do_query_helper(dbpedia_endpoint, query)




import readchar

def multiple_choice_selection_list(text_prompt, *options):
    """
    Allow selection of 0, 1, or more items from a selection list
    """

    selection_items = {
        str(key): item
        for key, item in enumerate(options, 1)
    }

    prompt = text_prompt + '\nHit enter or return key when done:\n' + \
              '\n'.join(f' {key}) {item}' for key, item in selection_items.items()) + \
              '\n[' + '/'.join(selection_items.keys()) + ']? :'

    ret = {}
    print(prompt)
    while True:
        key = readchar.readkey()
        name = selection_items.get(key)
        if name is not None:
            print(f"Selected '{name}'. Hit return key or enter to stop.")
            ret[key] = name
        else:
            return ret

def select_entities(people, places, organizations):
    "should return something like: \
    \
{'entities': ['Bill Gates || William Henry Gates III (born October 28, 1955) ' \
              'is an American busines...', \
              'Seattle || Washington (/ˈwɒʃɪŋtən/), officially the State of ' \
              'Washington, is a sta...', \
              'Apple || Hot Chocolate are a British soul band popular during ' \
              'the 1970s and 198...', \
              'Microsoft || Microsoft Corporation is an American multinational ' \
              'technology corporat...']} "

    ret = {'people': [], 'places': [], 'organizations': []}
    if len(people) > 0:
        ret['people'] = list(multiple_choice_selection_list("People", *people).values())
    if len(places) > 0:
        ret['places'] = list(multiple_choice_selection_list("Places", *places).values())
    if len(people) > 0:
        ret['organizations'] = list(multiple_choice_selection_list("Organizations", *organizations).values())
    return {'entities': ret['people'] + ret['places'] + ret['organizations']}


def get_query():
    return input('Enter a list of entities: ')




from .sparql import dbpedia_sparql
from .colorize import colorize_sparql

def flatten(a_list):
    return [item for items in a_list for item in items]

def dbpedia_get_relationships(s_uri, o_uri):
    query = (
        "SELECT DISTINCT ?p {{  {} ?p {} . FILTER (!regex(str(?p), 'wikiPage', 'i')) }} LIMIT 5"
        .format(s_uri, o_uri))
    results = dbpedia_sparql(query)
    print('Generated SPARQL to get relationships between two entities:')
    print(colorize_sparql(query))
    return [r for r in flatten(results) if not r == 'p']


def entity_results_to_relationship_links(uris):
    uris = [('<' + uri + '>') for uri in uris]
    relationship_statements = []
    for e1 in uris:
        for e2 in uris:
            if not e1 == e2:
                l1 = dbpedia_get_relationships(e1, e2)
                l2 = dbpedia_get_relationships(e2, e1)
                for x in l1:
                    relationship_statements.extend([[e1, e2, x]]) if not [e1,
                        e2, x] in relationship_statements else None
                for x in l2:
                    relationship_statements.extend([[e1, e2, x]]) if not [e1,
                        e2, x] in relationship_statements else None
    return relationship_statements




from .sparql import dbpedia_sparql
from .colorize import colorize_sparql

def dbpedia_get_entities_by_name(name, dbpedia_type):
    sparql = (
        'select distinct ?s ?comment {{ ?s ?p "{}"@en . ?s <http://www.w3.org/2000/01/rdf-schema#comment>  ?comment  . FILTER  (lang(?comment) = \'en\') . ?s <http://www.w3.org/1999/02/22-rdf-syntax-ns#type> {} . }} limit 15'
        .format(name, dbpedia_type))
    print('Generated SPARQL to get DBPedia entity URIs from a name:')
    print(colorize_sparql(sparql))
    return dbpedia_sparql(sparql)




from pprint import pprint
from .kgnutils import dbpedia_get_entities_by_name
from .textui import select_entities, get_query
from .relationships import entity_results_to_relationship_links
import spacy

try:
  nlp_model = spacy.load('en_core_web_sm')
except:
  print("Loading spaCy model file...")
  from os import system
  system("python -m spacy download en_core_web_sm")
  nlp_model = spacy.load('en_core_web_sm')


def entities_in_text(s):
    doc = nlp_model(s)
    ret = {}
    for [ename, etype] in [[entity.text, entity.label_] for entity in doc.ents]:
        if etype in ret:
            ret[etype] = ret[etype] + [ename]
        else:
            ret[etype] = [ename]
    return ret


entity_type_to_type_uri = {'PERSON': '<http://dbpedia.org/ontology/Person>',
    'GPE': '<http://dbpedia.org/ontology/Place>', 'ORG':
    '<http://dbpedia.org/ontology/Organisation>'}
short_comment_to_uri = {}


def shorten_comment(comment, uri):
    sc = comment[0:70:None] + '...'
    short_comment_to_uri[sc] = uri
    return sc


query = ''


def kgn():
    print("Knowledge Graph Navigator (note: only runs in a terminal)")
    while True:
        query = get_query()
        if query == 'quit' or query == 'q':
            break
        elist = entities_in_text(query)
        people_found_on_dbpedia = []
        places_found_on_dbpedia = []
        organizations_found_on_dbpedia = []
        global short_comment_to_uri
        short_comment_to_uri = {}
        for key in elist:
            type_uri = entity_type_to_type_uri[key]
            for name in elist[key]:
                dbp = dbpedia_get_entities_by_name(name, type_uri)
                for d in dbp:
                    short_comment = shorten_comment(d[1][1], d[0][1])
                    people_found_on_dbpedia.extend([name + ' || ' +
                        short_comment]) if key == 'PERSON' else None
                    places_found_on_dbpedia.extend([name + ' || ' +
                        short_comment]) if key == 'GPE' else None
                    organizations_found_on_dbpedia.extend([name + ' || ' +
                        short_comment]) if key == 'ORG' else None
        user_selected_entities = select_entities(people_found_on_dbpedia,
            places_found_on_dbpedia, organizations_found_on_dbpedia)
        uri_list = []
        for entity in user_selected_entities['entities']:
            short_comment = entity[4 + entity.index(' || '):None:None]
            uri_list.extend([short_comment_to_uri[short_comment]])
        print("\n\nEntity data:")
        pprint(user_selected_entities)
        print("\n\n")
        relation_data = (
            entity_results_to_relationship_links(uri_list))
        print('\n\nDiscovered relationship links:\n')
        for relationship in relation_data:
            print(relationship[0] + ' --> ' + relationship[2][1] + ' --> ' + relationship[1])

def kgn_console(query, use_gui):
    print("kgn_console is not implemented")





from sqlite3 import connect, version, Error
import json
from os import getenv
db_path = '{}/.kgn_hy_cache.db'.format(getenv('HOME'))


def create_db():
    try:
        conn = connect(db_path)
        print(version)
        cur = conn.cursor()
        cur.execute(
            'CREATE TABLE dbpedia (query string  PRIMARY KEY ASC, data json)')
        _hy_anon_var_1 = conn.close()
    except Exception as e:
        _hy_anon_var_1 = print(e)
    return _hy_anon_var_1


def save_query_results_dbpedia(query, result):
    try:
        conn = connect(db_path)
        cur = conn.cursor()
        cur.execute('insert into dbpedia (query, data) values (?, ?)', [
            query, json.dumps(result)])
        conn.commit()
        _hy_anon_var_3 = conn.close()
    except Exception as e:
        _hy_anon_var_3 = print(e)
    return _hy_anon_var_3


def fetch_result_dbpedia(query):
    results = []
    conn = connect(db_path)
    cur = conn.cursor()
    cur.execute('select data from dbpedia where query = ? limit 1', [query])
    d = cur.fetchall()
    if len(d) > 0:
        results = json.loads(d[0][0])
    conn.close()
    return results


create_db()




from io import StringIO

def blue(s):
    return '{}{}{}'.format('\x1b[94m', s, '\x1b[0m')

def red(s):
    return '{}{}{}'.format('\x1b[91m', s, '\x1b[0m')

def green(s):
    return '{}{}{}'.format('\x1b[92m', s, '\x1b[0m')

def pink(s):
    return '{}{}{}'.format('\x1b[95m', s, '\x1b[0m')

def bold(s):
    return '{}{}{}'.format('\x1b[1m', s, '\x1b[0m')

def tokenize_keep_uris(s):
    return s.split()

def colorize_sparql(s):
    tokens = tokenize_keep_uris(s.replace('{', ' { ').replace('}', ' } ').
        replace('.', ' . '))
    ret = StringIO()
    for token in tokens:
        (ret.write(red(token)) if token[0] == '?' else ret.write(blue(token
            )) if token in ['where', 'select', 'distinct', 'option',
            'filter', 'FILTER', 'OPTION', 'DISTINCT', 'SELECT', 'WHERE'] else
            ret.write(bold(token)) if token[0] == '<' else ret.write(token)
            ) if len(token) > 0 else None
        ret.write(' ') if not token == '?' else None
    ret.seek(0)
    return ret.read()




from setuptools import setup
import os

VERSION = "0.46"

def get_long_description():
    with open(
        os.path.join(os.path.dirname(os.path.abspath(__file__)), "README.md"),
        encoding="utf8",
    ) as fp:
        return fp.read()


setup(
    name="kgn",
    description="Interactive exploration of Knowledge Graphs using natural language",
    long_description=get_long_description(),
    long_description_content_type="text/markdown",
    author="Mark Watson",
    url="https://github.com/mark-watson/kgn",
    project_urls={
        "Issues": "https://github.com/mark-watson/kgn/issues",
        "CI": "https://github.com/mark-watson/kgn/actions",
        "Changelog": "https://github.com/mark-watson/kgn/releases",
    },
    license="Apache License, Version 2.0",
    version=VERSION,
    packages=["kgn"],
    entry_points="""
        [console_scripts]
        kgn=kgn.cli:cli
    """,
    install_requires=["click", "requests", "spacy", "readchar"],
    extras_require={
        "test": ["pytest"]
    },
    python_requires=">=3.7",
)
